package org.apache.bcel.generic;

package .var.folders.cl.9twcgjfx4vsg5zl73c6wxmvm0000gn.T;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import org.apache.bcel.classfile.AnnotationEntry;
import org.apache.bcel.classfile.Annotations;
import org.apache.bcel.classfile.Attribute;
import org.apache.bcel.classfile.Code;
import org.apache.bcel.classfile.CodeException;
import org.apache.bcel.classfile.ExceptionTable;
import org.apache.bcel.classfile.LineNumber;
import org.apache.bcel.classfile.LineNumberTable;
import org.apache.bcel.classfile.LocalVariable;
import org.apache.bcel.classfile.LocalVariableTable;
import org.apache.bcel.classfile.LocalVariableTypeTable;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.classfile.ParameterAnnotationEntry;
import org.apache.bcel.classfile.ParameterAnnotations;
import org.apache.bcel.classfile.Utility;
import org.apache.bcel.generic.AnnotationEntryGen;
import org.apache.bcel.generic.BranchInstruction;
import org.apache.bcel.generic.ClassGenException;
import org.apache.bcel.generic.CodeExceptionGen;
import org.apache.bcel.generic.ConstantPoolGen;
import org.apache.bcel.generic.FieldGenOrMethodGen;
import org.apache.bcel.generic.IndexedInstruction;
import org.apache.bcel.generic.Instruction;
import org.apache.bcel.generic.InstructionHandle;
import org.apache.bcel.generic.InstructionList;
import org.apache.bcel.generic.InstructionTargeter;
import org.apache.bcel.generic.LineNumberGen;
import org.apache.bcel.generic.LocalVariableGen;
import org.apache.bcel.generic.MethodObserver;
import org.apache.bcel.generic.ObjectType;
import org.apache.bcel.generic.Select;
import org.apache.bcel.generic.TargetLostException;
import org.apache.bcel.generic.Type;
import org.apache.bcel.generic.TypedInstruction;
import org.apache.bcel.util.BCELComparator;
import org.apache.commons.lang3.ArrayUtils;

public class MethodGen extends FieldGenOrMethodGen {
  private static BCELComparator bcelComparator = (BCELComparator)new Object();
  
  private String className;
  
  private Type[] argTypes;
  
  private String[] argNames;
  
  private int maxLocals;
  
  private int maxStack;
  
  private InstructionList il;
  
  private boolean stripAttributes;
  
  private LocalVariableTypeTable localVariableTypeTable;
  
  private static byte[] getByteCodes(Method method) {
    Code code = method.getCode();
    if (code == null)
      throw new IllegalStateException(String.format("The method '%s' has no code.", new Object[] { method })); 
    return code.getCode();
  }
  
  public static BCELComparator getComparator() {
    return bcelComparator;
  }
  
  public static int getMaxStack(ConstantPoolGen cp, InstructionList il, CodeExceptionGen[] et) {
    BranchStack branchTargets = new BranchStack();
    for (CodeExceptionGen element : et) {
      InstructionHandle handlerPc = element.getHandlerPC();
      if (handlerPc != null)
        branchTargets.push(handlerPc, 1); 
    } 
    int stackDepth = 0;
    int maxStackDepth = 0;
    InstructionHandle ih = il.getStart();
    while (ih != null) {
      Instruction instruction = ih.getInstruction();
      short opcode = instruction.getOpcode();
      int delta = instruction.produceStack(cp) - instruction.consumeStack(cp);
      stackDepth += delta;
      if (stackDepth > maxStackDepth)
        maxStackDepth = stackDepth; 
      if (instruction instanceof BranchInstruction) {
        BranchInstruction branch = (BranchInstruction)instruction;
        if (instruction instanceof Select) {
          Select select = (Select)branch;
          InstructionHandle[] targets = select.getTargets();
          for (InstructionHandle target : targets)
            branchTargets.push(target, stackDepth); 
          ih = null;
        } else if (!(branch instanceof org.apache.bcel.generic.IfInstruction)) {
          if (opcode == 168 || opcode == 201)
            branchTargets.push(ih.getNext(), stackDepth - 1); 
          ih = null;
        } 
        branchTargets.push(branch.getTarget(), stackDepth);
      } else if (opcode == 191 || opcode == 169 || (opcode >= 172 && opcode <= 177)) {
        ih = null;
      } 
      if (ih != null)
        ih = ih.getNext(); 
      if (ih == null) {
        BranchTarget bt = branchTargets.pop();
        if (bt != null) {
          ih = bt.target;
          stackDepth = bt.stackDepth;
        } 
      } 
    } 
    return maxStackDepth;
  }
  
  public static void setComparator(BCELComparator comparator) {
    bcelComparator = comparator;
  }
  
  private final List<LocalVariableGen> variableList = new ArrayList<>();
  
  private final List<LineNumberGen> lineNumberList = new ArrayList<>();
  
  private final List<CodeExceptionGen> exceptionList = new ArrayList<>();
  
  private final List<String> throwsList = new ArrayList<>();
  
  private final List<Attribute> codeAttrsList = new ArrayList<>();
  
  private List<AnnotationEntryGen>[] paramAnnotations;
  
  private boolean hasParameterAnnotations;
  
  private boolean haveUnpackedParameterAnnotations;
  
  private List<MethodObserver> observers;
  
  public MethodGen(int accessFlags, Type returnType, Type[] argTypes, String[] argNames, String methodName, String className, InstructionList il, ConstantPoolGen cp) {
    super(accessFlags);
    setType(returnType);
    setArgumentTypes(argTypes);
    setArgumentNames(argNames);
    setName(methodName);
    setClassName(className);
    setInstructionList(il);
    setConstantPool(cp);
    boolean abstract_ = (isAbstract() || isNative());
    InstructionHandle start = null;
    InstructionHandle end = null;
    if (!abstract_) {
      start = il.getStart();
      if (!isStatic() && className != null)
        addLocalVariable("this", ObjectType.getInstance(className), start, end); 
    } 
    if (argTypes != null) {
      int size = argTypes.length;
      for (Type argType : argTypes) {
        if (Type.VOID == argType)
          throw new ClassGenException("'void' is an illegal argument type for a method"); 
      } 
      if (argNames != null) {
        if (size != argNames.length)
          throw new ClassGenException("Mismatch in argument array lengths: " + size + " vs. " + argNames.length); 
      } else {
        argNames = new String[size];
        for (int i = 0; i < size; i++)
          argNames[i] = "arg" + i; 
        setArgumentNames(argNames);
      } 
      if (!abstract_)
        for (int i = 0; i < size; i++)
          addLocalVariable(argNames[i], argTypes[i], start, end);  
    } 
  }
  
  public MethodGen(Method method, String className, ConstantPoolGen cp) {
    this(method.getAccessFlags(), Type.getReturnType(method.getSignature()), Type.getArgumentTypes(method.getSignature()), (String[])null, method
        
        .getName(), className, 
        ((method.getAccessFlags() & 0x500) == 0) ? new InstructionList(getByteCodes(method)) : null, cp);
    Attribute[] attributes = method.getAttributes();
    for (Attribute attribute : attributes) {
      Attribute a = attribute;
      if (a instanceof Code) {
        Code c = (Code)a;
        setMaxStack(c.getMaxStack());
        setMaxLocals(c.getMaxLocals());
        CodeException[] ces = c.getExceptionTable();
        if (ces != null)
          for (CodeException ce : ces) {
            InstructionHandle end;
            int type = ce.getCatchType();
            ObjectType cType = null;
            if (type > 0) {
              String cen = method.getConstantPool().getConstantString(type, (byte)7);
              cType = ObjectType.getInstance(cen);
            } 
            int endPc = ce.getEndPC();
            int length = (getByteCodes(method)).length;
            if (length == endPc) {
              end = this.il.getEnd();
            } else {
              end = this.il.findHandle(endPc);
              end = end.getPrev();
            } 
            addExceptionHandler(this.il.findHandle(ce.getStartPC()), end, this.il.findHandle(ce.getHandlerPC()), cType);
          }  
        Attribute[] cAttributes = c.getAttributes();
        for (Attribute cAttribute : cAttributes) {
          a = cAttribute;
          if (a instanceof LineNumberTable) {
            ((LineNumberTable)a).forEach(l -> {
                  InstructionHandle ih = this.il.findHandle(l.getStartPC());
                  if (ih != null)
                    addLineNumber(ih, l.getLineNumber()); 
                });
          } else if (a instanceof LocalVariableTable) {
            updateLocalVariableTable((LocalVariableTable)a);
          } else if (a instanceof LocalVariableTypeTable) {
            this.localVariableTypeTable = (LocalVariableTypeTable)a.copy(cp.getConstantPool());
          } else {
            addCodeAttribute(a);
          } 
        } 
      } else if (a instanceof ExceptionTable) {
        Collections.addAll(this.throwsList, ((ExceptionTable)a).getExceptionNames());
      } else if (a instanceof Annotations) {
        Annotations runtimeAnnotations = (Annotations)a;
        runtimeAnnotations.forEach(element -> addAnnotationEntry(new AnnotationEntryGen(element, cp, false)));
      } else {
        addAttribute(a);
      } 
    } 
  }
  
  public void addAnnotationsAsAttribute(ConstantPoolGen cp) {
    addAll(AnnotationEntryGen.getAnnotationAttributes(cp, getAnnotationEntries()));
  }
  
  public void addCodeAttribute(Attribute a) {
    this.codeAttrsList.add(a);
  }
  
  public void addException(String className) {
    this.throwsList.add(className);
  }
  
  public CodeExceptionGen addExceptionHandler(InstructionHandle startPc, InstructionHandle endPc, InstructionHandle handlerPc, ObjectType catchType) {
    if (startPc == null || endPc == null || handlerPc == null)
      throw new ClassGenException("Exception handler target is null instruction"); 
    CodeExceptionGen c = new CodeExceptionGen(startPc, endPc, handlerPc, catchType);
    this.exceptionList.add(c);
    return c;
  }
  
  public LineNumberGen addLineNumber(InstructionHandle ih, int srcLine) {
    LineNumberGen l = new LineNumberGen(ih, srcLine);
    this.lineNumberList.add(l);
    return l;
  }
  
  public LocalVariableGen addLocalVariable(String name, Type type, InstructionHandle start, InstructionHandle end) {
    return addLocalVariable(name, type, this.maxLocals, start, end);
  }
  
  public LocalVariableGen addLocalVariable(String name, Type type, int slot, InstructionHandle start, InstructionHandle end) {
    return addLocalVariable(name, type, slot, start, end, slot);
  }
  
  public LocalVariableGen addLocalVariable(String name, Type type, int slot, InstructionHandle start, InstructionHandle end, int origIndex) {
    byte t = type.getType();
    if (t != 16) {
      int add = type.getSize();
      if (slot + add > this.maxLocals)
        this.maxLocals = slot + add; 
      LocalVariableGen l = new LocalVariableGen(slot, name, type, start, end, origIndex);
      int i;
      if ((i = this.variableList.indexOf(l)) >= 0) {
        this.variableList.set(i, l);
      } else {
        this.variableList.add(l);
      } 
      return l;
    } 
    throw new IllegalArgumentException("Can not use " + type + " as type for local variable");
  }
  
  public void addObserver(MethodObserver o) {
    if (this.observers == null)
      this.observers = new ArrayList<>(); 
    this.observers.add(o);
  }
  
  public void addParameterAnnotation(int parameterIndex, AnnotationEntryGen annotation) {
    ensureExistingParameterAnnotationsUnpacked();
    if (!this.hasParameterAnnotations) {
      List[] arrayOfList = new List[this.argTypes.length];
      this.paramAnnotations = (List<AnnotationEntryGen>[])arrayOfList;
      this.hasParameterAnnotations = true;
    } 
    List<AnnotationEntryGen> existingAnnotations = this.paramAnnotations[parameterIndex];
    if (existingAnnotations != null) {
      existingAnnotations.add(annotation);
    } else {
      List<AnnotationEntryGen> l = new ArrayList<>();
      l.add(annotation);
      this.paramAnnotations[parameterIndex] = l;
    } 
  }
  
  public void addParameterAnnotationsAsAttribute(ConstantPoolGen cp) {
    if (!this.hasParameterAnnotations)
      return; 
    Attribute[] attrs = AnnotationEntryGen.getParameterAnnotationAttributes(cp, this.paramAnnotations);
    if (attrs != null)
      addAll(attrs); 
  }
  
  private Attribute[] addRuntimeAnnotationsAsAttribute(ConstantPoolGen cp) {
    Attribute[] attrs = AnnotationEntryGen.getAnnotationAttributes(cp, getAnnotationEntries());
    addAll(attrs);
    return attrs;
  }
  
  private Attribute[] addRuntimeParameterAnnotationsAsAttribute(ConstantPoolGen cp) {
    if (!this.hasParameterAnnotations)
      return Attribute.EMPTY_ARRAY; 
    Attribute[] attrs = AnnotationEntryGen.getParameterAnnotationAttributes(cp, this.paramAnnotations);
    addAll(attrs);
    return attrs;
  }
  
  private void adjustLocalVariableTypeTable(LocalVariableTable lvt) {
    LocalVariable[] lv = lvt.getLocalVariableTable();
    for (LocalVariable element : this.localVariableTypeTable.getLocalVariableTypeTable()) {
      for (LocalVariable l : lv) {
        if (element.getName().equals(l.getName()) && element.getIndex() == l.getOrigIndex()) {
          element.setLength(l.getLength());
          element.setStartPC(l.getStartPC());
          element.setIndex(l.getIndex());
          break;
        } 
      } 
    } 
  }
  
  public org.apache.bcel.generic.MethodGen copy(String className, ConstantPoolGen cp) {
    Method m = ((org.apache.bcel.generic.MethodGen)clone()).getMethod();
    org.apache.bcel.generic.MethodGen mg = new org.apache.bcel.generic.MethodGen(m, className, getConstantPool());
    if (getConstantPool() != cp) {
      mg.setConstantPool(cp);
      mg.getInstructionList().replaceConstantPool(getConstantPool(), cp);
    } 
    return mg;
  }
  
  private void ensureExistingParameterAnnotationsUnpacked() {
    if (this.haveUnpackedParameterAnnotations)
      return; 
    Attribute[] attrs = getAttributes();
    ParameterAnnotations paramAnnVisAttr = null;
    ParameterAnnotations paramAnnInvisAttr = null;
    for (Attribute attribute : attrs) {
      if (attribute instanceof ParameterAnnotations) {
        if (!this.hasParameterAnnotations) {
          List[] arrayOfList = new List[this.argTypes.length];
          this.paramAnnotations = (List<AnnotationEntryGen>[])arrayOfList;
          Arrays.setAll(this.paramAnnotations, i -> new ArrayList());
        } 
        this.hasParameterAnnotations = true;
        ParameterAnnotations rpa = (ParameterAnnotations)attribute;
        if (rpa instanceof org.apache.bcel.classfile.RuntimeVisibleParameterAnnotations) {
          paramAnnVisAttr = rpa;
        } else {
          paramAnnInvisAttr = rpa;
        } 
        ParameterAnnotationEntry[] parameterAnnotationEntries = rpa.getParameterAnnotationEntries();
        for (int j = 0; j < parameterAnnotationEntries.length; j++) {
          ParameterAnnotationEntry immutableArray = rpa.getParameterAnnotationEntries()[j];
          List<AnnotationEntryGen> mutable = makeMutableVersion(immutableArray.getAnnotationEntries());
          this.paramAnnotations[j].addAll(mutable);
        } 
      } 
    } 
    if (paramAnnVisAttr != null)
      removeAttribute(paramAnnVisAttr); 
    if (paramAnnInvisAttr != null)
      removeAttribute(paramAnnInvisAttr); 
    this.haveUnpackedParameterAnnotations = true;
  }
  
  public boolean equals(Object obj) {
    return bcelComparator.equals(this, obj);
  }
  
  public List<AnnotationEntryGen> getAnnotationsOnParameter(int i) {
    ensureExistingParameterAnnotationsUnpacked();
    if (!this.hasParameterAnnotations || i > this.argTypes.length)
      return null; 
    return this.paramAnnotations[i];
  }
  
  public String getArgumentName(int i) {
    return this.argNames[i];
  }
  
  public String[] getArgumentNames() {
    return (String[])this.argNames.clone();
  }
  
  public Type getArgumentType(int i) {
    return this.argTypes[i];
  }
  
  public Type[] getArgumentTypes() {
    return (Type[])this.argTypes.clone();
  }
  
  public String getClassName() {
    return this.className;
  }
  
  public Attribute[] getCodeAttributes() {
    return this.codeAttrsList.<Attribute>toArray(Attribute.EMPTY_ARRAY);
  }
  
  private CodeException[] getCodeExceptions() {
    int size = this.exceptionList.size();
    CodeException[] cExc = new CodeException[size];
    Arrays.setAll(cExc, i -> ((CodeExceptionGen)this.exceptionList.get(i)).getCodeException(getConstantPool()));
    return cExc;
  }
  
  public CodeExceptionGen[] getExceptionHandlers() {
    return this.exceptionList.<CodeExceptionGen>toArray(CodeExceptionGen.EMPTY_ARRAY);
  }
  
  public String[] getExceptions() {
    return this.throwsList.<String>toArray(ArrayUtils.EMPTY_STRING_ARRAY);
  }
  
  private ExceptionTable getExceptionTable(ConstantPoolGen cp) {
    int size = this.throwsList.size();
    int[] ex = new int[size];
    Arrays.setAll(ex, i -> cp.addClass(this.throwsList.get(i)));
    return new ExceptionTable(cp.addUtf8("Exceptions"), 2 + 2 * size, ex, cp.getConstantPool());
  }
  
  public InstructionList getInstructionList() {
    return this.il;
  }
  
  public LineNumberGen[] getLineNumbers() {
    return this.lineNumberList.<LineNumberGen>toArray(LineNumberGen.EMPTY_ARRAY);
  }
  
  public LineNumberTable getLineNumberTable(ConstantPoolGen cp) {
    int size = this.lineNumberList.size();
    LineNumber[] ln = new LineNumber[size];
    Arrays.setAll(ln, i -> ((LineNumberGen)this.lineNumberList.get(i)).getLineNumber());
    return new LineNumberTable(cp.addUtf8("LineNumberTable"), 2 + ln.length * 4, ln, cp.getConstantPool());
  }
  
  public LocalVariableGen[] getLocalVariables() {
    int size = this.variableList.size();
    LocalVariableGen[] lg = new LocalVariableGen[size];
    this.variableList.toArray(lg);
    for (int i = 0; i < size; i++) {
      if (lg[i].getStart() == null && this.il != null)
        lg[i].setStart(this.il.getStart()); 
      if (lg[i].getEnd() == null && this.il != null)
        lg[i].setEnd(this.il.getEnd()); 
    } 
    if (size > 1)
      Arrays.sort(lg, Comparator.comparingInt(LocalVariableGen::getIndex)); 
    return lg;
  }
  
  public LocalVariableTable getLocalVariableTable(ConstantPoolGen cp) {
    LocalVariableGen[] lg = getLocalVariables();
    int size = lg.length;
    LocalVariable[] lv = new LocalVariable[size];
    Arrays.setAll(lv, i -> lg[i].getLocalVariable(cp));
    return new LocalVariableTable(cp.addUtf8("LocalVariableTable"), 2 + lv.length * 10, lv, cp.getConstantPool());
  }
  
  public LocalVariableTypeTable getLocalVariableTypeTable() {
    return this.localVariableTypeTable;
  }
  
  public int getMaxLocals() {
    return this.maxLocals;
  }
  
  public int getMaxStack() {
    return this.maxStack;
  }
  
  public Method getMethod() {
    String signature = getSignature();
    ConstantPoolGen cp = getConstantPool();
    int nameIndex = cp.addUtf8(getName());
    int signatureIndex = cp.addUtf8(signature);
    byte[] byteCode = (this.il != null) ? this.il.getByteCode() : null;
    LineNumberTable lnt = null;
    LocalVariableTable lvt = null;
    if (!this.variableList.isEmpty() && !this.stripAttributes) {
      updateLocalVariableTable(getLocalVariableTable(cp));
      addCodeAttribute(lvt = getLocalVariableTable(cp));
    } 
    if (this.localVariableTypeTable != null) {
      if (lvt != null)
        adjustLocalVariableTypeTable(lvt); 
      addCodeAttribute(this.localVariableTypeTable);
    } 
    if (!this.lineNumberList.isEmpty() && !this.stripAttributes)
      addCodeAttribute(lnt = getLineNumberTable(cp)); 
    Attribute[] codeAttrs = getCodeAttributes();
    int attrsLen = 0;
    for (Attribute codeAttr : codeAttrs)
      attrsLen += codeAttr.getLength() + 6; 
    CodeException[] cExc = getCodeExceptions();
    int excLen = cExc.length * 8;
    Code code = null;
    if (byteCode != null && !isAbstract() && !isNative()) {
      Attribute[] attributes = getAttributes();
      for (Attribute a : attributes) {
        if (a instanceof Code)
          removeAttribute(a); 
      } 
      code = new Code(cp.addUtf8("Code"), 8 + byteCode.length + 2 + excLen + 2 + attrsLen, this.maxStack, this.maxLocals, byteCode, cExc, codeAttrs, cp.getConstantPool());
      addAttribute(code);
    } 
    Attribute[] annotations = addRuntimeAnnotationsAsAttribute(cp);
    Attribute[] parameterAnnotations = addRuntimeParameterAnnotationsAsAttribute(cp);
    ExceptionTable et = null;
    if (!this.throwsList.isEmpty())
      addAttribute(et = getExceptionTable(cp)); 
    Method m = new Method(getAccessFlags(), nameIndex, signatureIndex, getAttributes(), cp.getConstantPool());
    if (lvt != null)
      removeCodeAttribute(lvt); 
    if (this.localVariableTypeTable != null)
      removeCodeAttribute(this.localVariableTypeTable); 
    if (lnt != null)
      removeCodeAttribute(lnt); 
    if (code != null)
      removeAttribute(code); 
    if (et != null)
      removeAttribute(et); 
    removeRuntimeAttributes(annotations);
    removeRuntimeAttributes(parameterAnnotations);
    return m;
  }
  
  public Type getReturnType() {
    return getType();
  }
  
  public String getSignature() {
    return Type.getMethodSignature(getType(), this.argTypes);
  }
  
  public int hashCode() {
    return bcelComparator.hashCode(this);
  }
  
  private List<AnnotationEntryGen> makeMutableVersion(AnnotationEntry[] mutableArray) {
    List<AnnotationEntryGen> result = new ArrayList<>();
    for (AnnotationEntry element : mutableArray)
      result.add(new AnnotationEntryGen(element, getConstantPool(), false)); 
    return result;
  }
  
  public void removeCodeAttribute(Attribute a) {
    this.codeAttrsList.remove(a);
  }
  
  public void removeCodeAttributes() {
    this.localVariableTypeTable = null;
    this.codeAttrsList.clear();
  }
  
  public void removeException(String c) {
    this.throwsList.remove(c);
  }
  
  public void removeExceptionHandler(CodeExceptionGen c) {
    this.exceptionList.remove(c);
  }
  
  public void removeExceptionHandlers() {
    this.exceptionList.clear();
  }
  
  public void removeExceptions() {
    this.throwsList.clear();
  }
  
  public void removeLineNumber(LineNumberGen l) {
    this.lineNumberList.remove(l);
  }
  
  public void removeLineNumbers() {
    this.lineNumberList.clear();
  }
  
  public void removeLocalVariable(LocalVariableGen l) {
    l.dispose();
    this.variableList.remove(l);
  }
  
  public void removeLocalVariables() {
    this.variableList.forEach(LocalVariableGen::dispose);
    this.variableList.clear();
  }
  
  public void removeLocalVariableTypeTable() {
    this.localVariableTypeTable = null;
  }
  
  public void removeNOPs() {
    if (this.il != null)
      for (InstructionHandle ih = this.il.getStart(); ih != null; ih = next) {
        InstructionHandle next = ih.getNext();
        if (next != null && ih.getInstruction() instanceof org.apache.bcel.generic.NOP)
          try {
            this.il.delete(ih);
          } catch (TargetLostException e) {
            for (InstructionHandle target : e.getTargets()) {
              for (InstructionTargeter targeter : target.getTargeters())
                targeter.updateTarget(target, next); 
            } 
          }  
      }  
  }
  
  public void removeObserver(MethodObserver o) {
    if (this.observers != null)
      this.observers.remove(o); 
  }
  
  public void removeRuntimeAttributes(Attribute[] attrs) {
    for (Attribute attr : attrs)
      removeAttribute(attr); 
  }
  
  public void setArgumentName(int i, String name) {
    this.argNames[i] = name;
  }
  
  public void setArgumentNames(String[] argNames) {
    this.argNames = argNames;
  }
  
  public void setArgumentType(int i, Type type) {
    this.argTypes[i] = type;
  }
  
  public void setArgumentTypes(Type[] argTypes) {
    this.argTypes = argTypes;
  }
  
  public void setClassName(String className) {
    this.className = className;
  }
  
  public void setInstructionList(InstructionList il) {
    this.il = il;
  }
  
  public void setMaxLocals() {
    if (this.il != null) {
      int max = isStatic() ? 0 : 1;
      if (this.argTypes != null)
        for (Type argType : this.argTypes)
          max += argType.getSize();  
      for (InstructionHandle ih = this.il.getStart(); ih != null; ih = ih.getNext()) {
        Instruction ins = ih.getInstruction();
        if (ins instanceof org.apache.bcel.generic.LocalVariableInstruction || ins instanceof org.apache.bcel.generic.RET || ins instanceof org.apache.bcel.generic.IINC) {
          int index = ((IndexedInstruction)ins).getIndex() + ((TypedInstruction)ins).getType(getConstantPool()).getSize();
          if (index > max)
            max = index; 
        } 
      } 
      this.maxLocals = max;
    } else {
      this.maxLocals = 0;
    } 
  }
  
  public void setMaxLocals(int m) {
    this.maxLocals = m;
  }
  
  public void setMaxStack() {
    if (this.il != null) {
      this.maxStack = getMaxStack(getConstantPool(), this.il, getExceptionHandlers());
    } else {
      this.maxStack = 0;
    } 
  }
  
  public void setMaxStack(int m) {
    this.maxStack = m;
  }
  
  public void setReturnType(Type returnType) {
    setType(returnType);
  }
  
  public void stripAttributes(boolean flag) {
    this.stripAttributes = flag;
  }
  
  public final String toString() {
    String access = Utility.accessToString(getAccessFlags());
    String signature = Type.getMethodSignature(getType(), this.argTypes);
    signature = Utility.methodSignatureToString(signature, getName(), access, true, getLocalVariableTable(getConstantPool()));
    StringBuilder buf = new StringBuilder(signature);
    for (Attribute a : getAttributes()) {
      if (!(a instanceof Code) && !(a instanceof ExceptionTable))
        buf.append(" [").append(a).append("]"); 
    } 
    if (!this.throwsList.isEmpty())
      for (String throwsDescriptor : this.throwsList)
        buf.append("\n\t\tthrows ").append(throwsDescriptor);  
    return buf.toString();
  }
  
  public void update() {
    if (this.observers != null)
      for (MethodObserver observer : this.observers)
        observer.notify(this);  
  }
  
  private void updateLocalVariableTable(LocalVariableTable a) {
    removeLocalVariables();
    for (LocalVariable l : a.getLocalVariableTable()) {
      InstructionHandle start = this.il.findHandle(l.getStartPC());
      InstructionHandle end = this.il.findHandle(l.getStartPC() + l.getLength());
      if (null == start)
        start = this.il.getStart(); 
      addLocalVariable(l.getName(), Type.getType(l.getSignature()), l.getIndex(), start, end, l.getOrigIndex());
    } 
  }
}
